const exec = require('child_process').exec;
const events = require('events');
const eventEmitter = new events.EventEmitter();

// avail. info
// https://docs.mongodb.com/manual/reference/command/serverStatus/

/**
 * Create mongo command from config options
 * @param {object} config
 * @return {string} mongo command to be executed
 */
function createCmd(config) {
    var mongoCmd = 'mongo ' + config.dbPath;

    if (config.dbOptions.server.ssl) {
        mongoCmd += ' --ssl';

        if (!config.dbOptions.server.sslValidate) {
            mongoCmd += ' --sslAllowInvalidCertificates';
        }
    }

    if (config.dbOptions.user.length) mongoCmd += ' -u ' + config.dbOptions.user;
    if (config.dbOptions.pass.length) mongoCmd += ' -p ' + config.dbOptions.pass;

    // append mongo cmd and pipe to sed to filter useless stuff
    // don't display certain blocks as they contain arbitrary content that might break JSON sanitization
    mongoCmd += ' --eval "db.serverStatus({security: 0, tcmalloc: 0})"';
    mongoCmd += " | sed '/^MongoDB/d' | sed '/^connecting to/d' | sed '/W NETWORK/d'";

    return mongoCmd;
}

/**
 * parse mongo command output and return obj.
 * @param {string} stdin
 * @return {object} mongo result as an object
 */
function parseMongoRes(stdout) {
    // convert output into valid json
    var jsonReady = '';
    var regex = /("[A-Za-z 0-9:(){}<>\/.,;_-]+"| : |true|false|,|[0-9]+(?!\)))|(NumberLong\("[0-9]+"\))|(ISODate\(".*"\))|([a-zA-Z0-9()]+)/g;

    jsonReady += stdout.replace(regex, function (match, clean, quotedNumber, isoDate, toQuote) {
        if (toQuote) return `"${toQuote}"`;

        // sequences with quoted numbers inside are quoted
        // and existing quotes around numbers are escaped
        else if (quotedNumber) {
            return `"${quotedNumber.replace(/["']/g, '\\"')}"`;
        }

        // same for isoDate as the date is quoted ex: ISODate("....")
        else if (isoDate) {
            return `"${isoDate.replace(/["']/g, '\\"')}"`;
        }

        else return match;
    });

    return JSON.parse(jsonReady);
}

/**
 * get mongo information and cache the result
 * @callback – mongo result
 */
var mongoStats = function (config) {
    var cmdOutput;
    var ready = false;
    var mongoCmd = createCmd(config);

    exec(mongoCmd, (err, stdout, stderr) => {
        if (err) console.error('command error: ' + err);
        if (stderr) console.error('command error: ' + stderr);
        if (err || stderr) process.exit(3);

        cmdOutput = parseMongoRes(stdout);

        if (cmdOutput.errmsg) return console.log(cmdOutput.errmsg);

        ready = true;
        eventEmitter.emit('ready');
    });

    return (callback) => {
        if (ready) return callback(cmdOutput);

        // if function is not ready yet
        // add a listener that will call cb when ready
        eventEmitter.on('ready', () => {
            callback(cmdOutput);
        });
    };
};

/**
 * Initialize mongoStats with config
 * @param {object} confObject – contains mongoDB connection info
 */
function init(confObject) {
    var config = confObject;
    mongoStats = mongoStats(config);
}
exports.init = init;

function rawData(callback) {
    mongoStats((result) => {
        callback(result);
    });
}
exports.rawData = rawData;

function status(callback) {
    mongoStats((result) => {
        var status = (result.ok === 1) ? true : false;
        callback(status);
    });
}
exports.status = status;

function storageEngine(callback) {
    mongoStats((result) => {
        callback(result.storageEngine.name);
    });
}
exports.storageEngine = storageEngine;

function uptime(callback) {
    mongoStats((result) => {
        callback(result.uptime);
    });
}
exports.uptime = uptime;

function flushAvgTime(callback) {
    mongoStats((result) => {
        if (result.storageEngine.name === 'wiredTiger') return callback('NA');
        callback(result.backgroundFlushing.average_ms);
    });
}
exports.flushAvgTime = flushAvgTime;

function lastFlushTime(callback) {
    mongoStats((result) => {
        if (result.storageEngine.name === 'wiredTiger') return callback('NA');
        callback(result.backgroundFlushing.last_ms);
    });
}
exports.lastFlushTime = lastFlushTime;

function usedConnections(callback) {
    mongoStats((result) => {
        callback(result.connections.current * (100 / result.connections.available));
    });
}
exports.usedConnections = usedConnections;

function memory(callback) {
    mongoStats((result) => {
        callback({
            RAM: result.mem.resident,
            virtual: result.mem.virtual
        });
    });
}
exports.memory = memory;
